// MySQLClient.cpp : Defines the entry point for the console application.


#include "stdafx.h"
#include <iostream>
#include <fstream>
#include <my_global.h>
#include <my_sys.h>
#include <m_string.h>
#include <mysql.h>
#include <errmsg.h>
#include <my_getopt.h>
#include <m_ctype.h>
#include <my_dir.h>
#include <string>

//label definitions for trace segmentation
#define DRY  0			//raw unlabeled data point
#define WET  1			//data point indicates Wet flow

#define MIN_MEAS -1     //minimum level below which meas is labeled as OUTlier
#define MIN_DRY  2      // level at which we declare a point to be WET
#define EVENT_TRIGGER  2     // std at which we declare WET

//definitions for overflow events
#define NORMAL 0					//nothing detected (YET) input DRY and output DRY
#define ABNORMAL 1					// sensor state problem
#define OUTLIER  2

#define WET_OVERFLOW 1				// wet overflow event detected - input WET and output WET
#define DRY_OVERFLOW 2				// dry overflow event detected - input DRY and output WET
#define OVERFLOW_ANOMALY 3 //input WET and output DRY

#define WINDOW_LENGTH 150		//length of window used in computing sensor stats.
#define TEMPLATE_LENGTH 10		//length of edge detector filter
#define MAX_NUM_EVENTS 100
#define EXPECTED_SENSOR_STD 0.1       //std dev of sensor noise (expected)


//MySQL global names for EmNet's Database
static char			*opt_host_name			= "emnet3.homeip.net";  
static char			*opt_user_name			= "wwtp";
static char			*opt_password			= "cso07net";
static char			*opt_socket_name		= NULL;
static unsigned int opt_port_num			= 0;     
static char			*opt_db_name			= "CSO_CITY_3";
static unsigned int opt_flags				= 0;      

//pointer to connection handle and stored data
static MYSQL		*conn;                       
MYSQL_RES			*res_set;
bool verbose = FALSE;


//structures for storing data
//Node structure containing information about specific sensor node
struct Node{
	char Name[10];
	int  Sensor;
	int  ID;
	double scale;
	double offset;
};

//structure to hold date
struct Date{
	int Year;
	int Month;
	int Day;
	int Hour;
	int Minute;
	int Second;
	char DateString[20];
}; 

 
//structure to hold sensor data (currently scalar)
struct Data{
	unsigned long long length;
	double *time;
	double *meas;
	char Name[40];
	Date *date;
	int *type;
	int *sensor_state;
	double *mean;
	double *std;
	int num_events;
	int event_start_index[MAX_NUM_EVENTS];
	int event_stop_index[MAX_NUM_EVENTS];
};

 


int start_year = 2008;
int start_month = 11;
int start_day   = 25;
int start_hour  = 0;
int start_minute = 0;
int start_second = 0;

int stop_year = 2009;
int stop_month = 6;
int stop_day   = 25;
int stop_hour  = 0;
int stop_minute = 0;
int stop_second = 0;


//global variables used in program
Node InputNode;
Node OutputNode;
Date StartDate;
Date StopDate;
Data InputMeas;
Data OutputMeas; 

using namespace std;

//	Basic functions encapsulating opening and closing of EmNet's Database
//   void Open_CSOnet_DBserver(void)   =  opens data base and assigns to global conn
//	 void Close_CSOnet_DBserver(void)  =  closes an open conn handle

void Open_CSOnet_DBserver(void)
{
	conn = mysql_init(NULL);
	if (mysql_real_connect (conn, opt_host_name, opt_user_name, opt_password,
			  opt_db_name, opt_port_num, opt_socket_name, opt_flags) == NULL)
	{
		fprintf(stderr,"Intialize_InputNode:mysql_real_connect failed\n");
		mysql_close(conn);
		exit(1);
	}
}

void Close_CSOnet_DBserver(void)
{
	mysql_close(conn);
}

//functions that encapsulate working with INode structure
//	void Initialize_INode(Node *INode, char Name[10],int Sensor)
//		initialize Inode structure 
//		INode is handle to global INode structure
//		Name[10] is the node's name (e.g. G.21.1)
//		Sensor   is the number of the sensor attached to the Gnode
//
//  void print_INode (Node *INode)
//		print out the Node data so we can double check it
//

void Initialize_INode(Node *INode,char Name[10],int Sensor)
{
	MYSQL_ROW	row;
	unsigned int i;
	char sql_query[350];
	char s1[60];

	//initialize sensor number and node's name
	INode->Sensor = Sensor;
	strcpy_s(INode->Name,Name);
	
	//query the database to get the ID number of the named node
	//then store the ID number in the structure
	strcpy_s(sql_query,"SELECT ID FROM inodes where Name ='");
	strcat_s(sql_query,INode->Name);
	strcat_s(sql_query,"'");
	if (verbose) cout << "mysql_query:" << sql_query << endl;
	if (mysql_query(conn, sql_query)!=0)
	{
		printf("SELECT failed to return InputNode ID\n");
		mysql_close(conn);
		exit(1);
	}else{
		res_set = mysql_store_result(conn);
		while ((row = mysql_fetch_row(res_set))!=NULL)
		{
			for (i=0;i<mysql_num_fields(res_set);i++)
			{
				INode->ID = atoi(row[i]);
			}
		}
	}
	mysql_free_result(res_set);


	//query the database to get the scale and offset factors for the named node
	//then store this calibration data in the structure.
	strcpy_s(sql_query,"SELECT ");
	sprintf_s(s1,"a%i, b%i",InputNode.Sensor,InputNode.Sensor);
	strcat_s(sql_query,s1);
	strcat_s(sql_query," FROM inodes WHERE ID=");
	sprintf_s(s1,"%i",InputNode.ID);
	strcat_s(sql_query,s1);
	if (verbose) cout << "sql_query: " << sql_query << endl;
	
	if (mysql_query(conn,sql_query)!=0)
	{
		printf("Select failed to return InputNode Scale\n");
		mysql_close(conn);
		exit(1);
	}else{
		res_set = mysql_store_result(conn);
		row = mysql_fetch_row(res_set);
		INode->scale = atof(row[0]);
		INode->offset = atof(row[1]);
	}
	mysql_free_result(res_set);
}

void print_INode (Node *INode)
{
	cout << INode->Name << "(S" << INode->Sensor << ")="<< INode->ID;
	cout << "	scale factor (a) =" << INode->scale << "	offset (b) =" << INode->offset << endl;
}

//functions that facilitate working with dates
//  void Initialize_Date(Date *DDate, int Y, int M, int D, int H, int M, int S)
//		initializes Date structure with specific date (Y-M-D) and time (H-M-S)
//		creates string that is used by MySQL queries
//
//	void print_Date(Date *DDate)
//		prints DateString in the structure
 
void Initialize_Date(Date *DDate, int Year, int Month, int Day, int Hour, int Minute, int Second)
{
	DDate->Year  = Year;
	DDate->Month = Month;
	DDate->Day   = Day;
	DDate->Hour  = Hour;
	DDate->Minute= Minute;
	DDate->Second= Second;
	sprintf_s(DDate->DateString,"%i-%i-%i %i:%i:%i",Year,Month,Day,Hour,Minute,Second);
}

void Initialize_Date2(Date *date, char *timestring)
{
	char t[25];
	char *year;
	char *month;
	char *day;
	char *hour;
	char *minute;
	char *second;
	char delim[] = " -:";

	strcpy_s(t,timestring);
	year = strtok(t,delim);
	date->Year = atoi(year);

	strcpy_s(t,t+strlen(year)+1);
	month = strtok(t,delim);
	date->Month = atoi(month);

	strcpy_s(t,t+strlen(month)+1);
	day = strtok(t,delim);
	date->Day = atoi(day);

	strcpy_s(t,t+strlen(day)+1);
	hour = strtok(t,delim);
	date->Hour = atoi(hour);

	strcpy_s(t,t+strlen(hour)+1);
	minute = strtok(t,delim);
	date->Minute = atoi(minute);

	strcpy_s(t,t+strlen(minute)+1);
	second = strtok(t,delim);
	date->Second = atoi(second);

	sprintf_s(date->DateString,"%i-%i-%i %i:%i:%i",date->Year,date->Month,date->Day,date->Hour,date->Minute,date->Second);
}

void print_Date(Date *DDate)
{
	cout << DDate->DateString << endl;
	cout << DDate->Year << "-" << DDate->Month << "-" <<DDate->Day ;
	cout << " " << DDate->Hour << ":" << DDate->Minute << ":" << DDate->Second << endl;
}

//functions encapsulating working with data arrays
//
//	double fetch-time(char *timestring)
//		utility function that takes the DateString in a Date structure
//		and extracts the time in (seconds) as a double.
//
//   void print_SensorData(Data *sdata)
//		utility function to print the time and sensor data (scalar)
//		in the Data structure sdata.
//
//	void write_SensorData(Data *sdata, char *name)
//		utility function to write time and sensor data to a txt file
//		sensor data to be written is in sdata
//		output file name is char *name
//
//    void fetch_SensorData(Data *sdata, Node *INode, Date *start, Date *stop)
//		queries MySQL database for data in INode between given start and stop times.
//		returned data is stored in the Data structure sdata.


double fetch_time(char *timestring)
{
	double time;
	char t[25];
	char *year;
	char *month;
	char *day;
	char *hour;
	char *minute;
	char *second;
	char delim[] = " -:";

	strcpy_s(t,timestring);
	year = strtok(t,delim);
	time = 365*24*3600*(atoi(year)-2008);

	strcpy_s(t,t+strlen(year)+1);
	month = strtok(t,delim);
	time = time + 3600*24*30*atoi(month);

	strcpy_s(t,t+strlen(month)+1);
	day = strtok(t,delim);
	time = time + 3600*24*atoi(day);
 
	strcpy_s(t,t+strlen(day)+1);
	hour = strtok(t,delim);
	time = time+3600*atoi(hour);

	strcpy_s(t,t+strlen(hour)+1);
	minute = strtok(t,delim);
	time = time + 60*atoi(minute);

	strcpy_s(t,t+strlen(minute)+1);
	second = t;
	time = time + atoi(second);

	return time;
}




void print_SensorData(Data *sdata)
{
	unsigned long i,length;
	length = (unsigned long) sdata->length;
	cout << "Sensor Data for INode " << sdata->Name << endl;
	cout << "Length of Data Record = " << sdata->length<< endl;
	for (i=0;i<length;i++) cout << sdata->time[i] << " " << sdata->meas[i] << endl;
	cout << endl;
}

void write_SensorData(Data *sdata,char *name){
	unsigned long i, length;
	Date *date;
	
	length = (unsigned long) sdata->length;
	ofstream outFile(name,ios::out);
	for (i=0;i<length;i++)
	{
		date = &sdata->date[i];
		outFile << sdata->time[i];
		outFile << " " << sdata->meas[i] ;
		outFile << " " << sdata->type[i] ;
		outFile << " " << sdata->sensor_state[i];
		outFile << " " << sdata->mean[i];
		outFile << " " << sdata->std[i];
		outFile<< " " << date->DateString;
		outFile << endl;
	}
	outFile.close();
}

void fetch_SensorData(Data *sdata, Node *INode, Date *start, Date *stop)
{
	MYSQL_ROW row;
	unsigned int i,j;
	unsigned long length;
	char sql_query[350];
	char s1[60];
	double meas;

	//form query for sensor data
	strcpy_s(sdata->Name,INode->Name);
	strcat_s(sdata->Name,"(S");
	sprintf_s(s1,"%i",INode->Sensor);
	strcat_s(sdata->Name,s1);
	strcat_s(sdata->Name,")");

	strcpy_s(sql_query,"SELECT Time, Sens");
	sprintf_s(s1,"%i ",INode->Sensor);
	strcat_s(sql_query,s1);
	strcat_s(sql_query," FROM inodes_data WHERE I_ID=");
	sprintf_s(s1,"%i ",INode->ID);
	strcat_s(sql_query,s1);
	strcat_s(sql_query,"and Time between STR_TO_DATE('");
	strcat_s(sql_query,start->DateString);
	strcat_s(sql_query,"','%Y-%m-%d %T') and STR_TO_DATE('");
	strcat_s(sql_query,stop->DateString);
	strcat_s(sql_query,"','%Y-%m-%d %T')");

	if (verbose) cout << "sql_query:" << sql_query << endl;

	//query database for data.  store data in res_set
	//    take fetched data in res_set and store in Data structure
	if(mysql_query(conn, sql_query)!=0)
	{
		printf ("FETCH of Data failed\n");
	}else{
		res_set = mysql_store_result(conn);

		length = (unsigned long) mysql_num_rows(res_set);
		sdata->length = (long) length;
		sdata->meas = (double *) malloc(length*sizeof(double));
		sdata->time = (double *) malloc(length*sizeof(double));
		sdata->type = (int *) malloc(length*sizeof(int));
		sdata->date = (Date *) malloc(length*sizeof(Date));
		sdata->mean = (double *) malloc(length*sizeof(double));
		sdata->std  = (double *) malloc(length*sizeof(double));
		sdata->sensor_state = (int *) malloc(length*sizeof(int));
 
		for (i=0;i < length; i++)
		{
			sdata->type[i] = DRY;
			sdata->mean[i]     = 0.0;
			sdata->std[i]      = 0.0;
			sdata->sensor_state[i] = NORMAL;
			
			row = mysql_fetch_row(res_set);
			if (row==NULL) break;

			for (j=0;j < mysql_num_fields (res_set);j++)
			{
				if (j==0) 
				{
					sdata->time[i] = fetch_time(row[j]);
					Initialize_Date2(&sdata->date[i],row[j]);
				}
				else sdata->meas[i] = INode->scale * atoi(row[j]) + INode->offset;
			}
		}

		for (i=0;i<length;i++)
		{
			meas = sdata->meas[i];
	//		sdata->type[i]=DRY;
			if (meas < MIN_MEAS) sdata->sensor_state[i] = OUTLIER;
	//		if ((meas > MIN_DRY)&&(sdata->sensor_state[i]==NORMAL)) sdata->type[i] = WET;
		}

 
		int nsamp;
		for (i=WINDOW_LENGTH;i < length-WINDOW_LENGTH;i++)
		{
			sdata->mean[i]=0;
			sdata->std[i] =0;
			nsamp = 0;
			for (j=i-WINDOW_LENGTH;j<i+WINDOW_LENGTH;j++)
			{
				if (sdata->sensor_state[i]==NORMAL)
				{
					sdata->mean[i] += sdata->meas[j];
					sdata->std[i]  += (sdata->meas[j])*(sdata->meas[j]);
					nsamp++;
				}
			}
			sdata->mean[i] = sdata->mean[i] / ((double) nsamp);
			sdata->std[i]  = sdata->std[i] / ((double) nsamp);
			sdata->std[i]  = max(0,sqrt(sdata->std[i]- (sdata->mean[i])*(sdata->mean[i])));
			if ((sdata->std[i] < EXPECTED_SENSOR_STD/10)&&(sdata->type[i]==DRY)&&(sdata->sensor_state[i]!=OUTLIER)) sdata->sensor_state[i] = ABNORMAL;
		}

		bool  event_active=FALSE;
		double peak_level=0.0;
		double start_level = 0.0;
		double threshold = 100.0;
		sdata->num_events = 0;

		event_active=FALSE;
		int duration = 0;
		for (i=TEMPLATE_LENGTH;i<length-TEMPLATE_LENGTH;i++)
		{
			double tmp=0.0;
			for (j=i-TEMPLATE_LENGTH;j<i;j++)   tmp -= sdata->std[j];
			for (j=i;j<i+TEMPLATE_LENGTH;j++)   tmp += sdata->std[j];

			
			switch (event_active)
			{
				case TRUE:
					if (peak_level < sdata->std[i]) peak_level = sdata->std[i];
					sdata->type[i] = WET;
					duration++;
					threshold = start_level + (peak_level-start_level)/10;
					
					if ((sdata->sensor_state[i]==NORMAL)&&(sdata->std[i] < threshold)&&(duration >5))
					{
//						cout << "	-" << i << ":" << sdata->mean[i] << ":" << threshold << endl;
						sdata->type[i]=DRY;
						event_active=FALSE;
//						sdata->event_stop_index[sdata->num_events]=i;
//						(sdata->num_events)++;
//						if (sdata->num_events > MAX_NUM_EVENTS) sdata->num_events = MAX_NUM_EVENTS;
						start_level=0.0;
						peak_level = 0.0;
						duration = 0;
					}
					break;

				case FALSE:
					sdata->type[i]=DRY;
					if (((sdata->std[i]>0.5)||(tmp > EVENT_TRIGGER))&&(sdata->sensor_state[i]==NORMAL))
					{
						sdata->type[i]=WET;
						event_active=TRUE;
//						sdata->event_start_index[sdata->num_events]=i;
//						cout   << i ;
						start_level = sdata->std[i-TEMPLATE_LENGTH];
						peak_level = sdata->std[i+TEMPLATE_LENGTH];
					}
					break;

				default:
					break;
			}

		}
		int past_state = DRY;
		sdata->num_events = 0;
		for (i=0;i<length;i++)
		{
			if ((sdata->type[i]==WET)&&(past_state==DRY)){
				sdata->event_start_index[sdata->num_events]=i;

//				cout << i;
			}

			if ((sdata->type[i]==DRY)&&(past_state==WET))
			{
				sdata->event_stop_index[sdata->num_events]=i;
				(sdata->num_events)++;
//				cout << "-" << i << endl;
			}
			past_state = sdata->type[i];
		}
	}
}


// functions to create data arrays containing overflow data traces
bool IsBefore(Date *dateStart, Date *dateStop)
{
	if (dateStart->Year > dateStop->Year) return FALSE;
	if ((dateStart->Year == dateStop->Year)&&(dateStart->Month > dateStop->Month)) return FALSE;
	if ((dateStart->Month == dateStop->Month)&&(dateStart->Day > dateStop->Day)) return FALSE;
	if ((dateStart->Day  == dateStop->Day)  &&(dateStart->Hour > dateStop->Hour)) return FALSE;
	if ((dateStart->Hour == dateStop->Hour) &&(dateStart->Minute > dateStop->Minute)) return FALSE;
	if ((dateStart->Minute == dateStop->Minute) && (dateStart->Second > dateStop->Second)) return FALSE;
	return TRUE;
}

 

long FindbyDate(Date *mydate,Data *sdata)
{
	long i,index;
	Date *Dateptr;

	index = 0;
	long length = (long) sdata->length;
	//cout << "length = " << length << endl;
	for (i=0;i<length;i++)
	{
		Dateptr = &(sdata->date[i]);
		if (IsBefore( Dateptr, mydate)==TRUE)
		{
			index = i;
			//cout << i << ":" << mydate->DateString << " > " << Dateptr->DateString <<endl;
		}
	}
	//cout << index << " - " << (sdata->date[index]).DateString << endl;
	return index;
}

void identify_overflows(Data *sdataIn, Data *sdataOut)
{
	int i,j;
	long lenIn,lenOut;
	long eventStopIn,eventStartIn;
	long eventStopOut,eventStartOut;
 
	Date *startDateInptr;
	Date *stopDateInptr;
	Date *startDateOutptr;
	Date *stopDateOutptr;
	Date *date;

 

	lenIn = sdataIn->num_events;
	lenOut = sdataOut->num_events;

	cout << "CSO OVERFLOW SUMARY" << endl << endl;
	cout << "	" << lenIn << " Input Events and " << lenOut << " Output Events" << endl;

	for (i=0;i<lenOut;i++)
	{
		eventStartOut = sdataOut->event_start_index[i];
		eventStopOut  = sdataOut->event_stop_index[i];

		startDateOutptr = &(sdataOut->date[eventStartOut]);
		stopDateOutptr  = &(sdataOut->date[eventStopOut]);
		eventStartIn  = FindbyDate(startDateOutptr,sdataIn);
		eventStopIn   = FindbyDate(stopDateOutptr,sdataIn);

		cout << "CSO EVENT " << i << endl;
		cout << "	" << "duration = " << eventStopOut-eventStartOut << " samples" << endl;
		cout << "	" << "starting " << startDateOutptr->DateString << endl;
		cout << "	" << "ending   " << stopDateOutptr->DateString << endl;
		cout << endl;

 

		char name[25];
		char s1[10];

		strcpy_s(name,"data/CSO_resultOUT");
		sprintf_s(s1,"%i",i);
		strcat_s(name,s1);
		strcat_s(name,".txt");
		ofstream outFileOut(name, ios::out);
		for (j=eventStartOut;j<eventStopOut;j++)
		{
			date = &sdataOut->date[j];
			outFileOut << sdataOut->time[j];
			outFileOut << " " << sdataOut->meas[j];
			outFileOut << " " << sdataOut->std[j];
			outFileOut << " " << sdataOut->type[j];
			outFileOut << " " << date->DateString;
			outFileOut << endl;
		}
		outFileOut.close();


		strcpy_s(name,"data/CSO_resultIN");
		sprintf_s(s1,"%i",i);
		strcat_s(name,s1);
		strcat_s(name,".txt");
		ofstream outFileIn(name,ios::out);

		for (j=eventStartIn;j<eventStopIn;j++)
		{
			date = &sdataIn->date[j];
			outFileIn << sdataIn->time[j];
			outFileIn << " " << sdataIn->meas[j];
			outFileIn << " " << sdataIn->std[j];
			outFileIn << " " << sdataIn->type[j];
			outFileIn << " " << date->DateString;
			outFileIn << endl;
		}
		outFileIn.close();

		int overflow_type = DRY;
		for (j=0;j<lenIn;j++)
		{
			eventStartIn = sdataIn->event_start_index[j];
			eventStopIn  = sdataIn->event_stop_index[j];
			startDateInptr = &(sdataIn->date[eventStartIn]);
			stopDateInptr  = &(sdataIn->date[eventStopIn]);
			if ((IsBefore(startDateOutptr,stopDateInptr))&&(IsBefore(startDateInptr,stopDateOutptr)))
			{
				overflow_type = WET;
				break;
			}
		}
		if (overflow_type==WET)
		{
			cout << "	WET cso event " << endl;
		}
		else
		{
			cout << "	DRY cso event " << endl;
		}
		cout << endl;


		 
		
	}

	cout << "TOTAL NUMBER OF CSO EVENTS = " << lenOut << endl;


}


//main body of program
//   right now this is programmed to use the simple 2-node scenario provdied by EmNET
//   in which G.21.1(S3) is the input and G.21.1(S1) is the output

int _tmain(int argc, _TCHAR* argv[])
{

	Open_CSOnet_DBserver();
  
	Initialize_INode(&InputNode,"G.21.1",3);
	if (verbose) print_INode(&InputNode);
 
	Initialize_INode(&OutputNode,"G.21.1",1);
	if (verbose) print_INode(&OutputNode);

	//Initialize_Date(&StartDate,2008,7,25,00,00,00);
	Initialize_Date(&StartDate,start_year,start_month,start_day,start_hour,start_minute,start_second);
	if (verbose) print_Date(&StartDate);

	//Initialize_Date(&StopDate,2009,7,25,00,00,00);
	Initialize_Date(&StopDate,stop_year,stop_month,stop_day,stop_hour,stop_minute,stop_second);

	if (verbose) print_Date(&StopDate);

	fetch_SensorData(&InputMeas, &InputNode, &StartDate, &StopDate);
	if (verbose) print_SensorData(&InputMeas);
	write_SensorData(&InputMeas,"data/InputMeas.txt");

	fetch_SensorData(&OutputMeas, &OutputNode, &StartDate, &StopDate);
	if (verbose) print_SensorData(&OutputMeas);
	write_SensorData(&OutputMeas,"data/OutputMeas.txt");

    
	identify_overflows(&InputMeas,&OutputMeas);
 
 

	Close_CSOnet_DBserver();

}

 